/*  VT1500 Device Driver for Linux, Version 1.0.10
 *  Copyright (C) 1995-2001  M. Gutschke
 *
 *  At the time of writing my e-mail address is:
 *	Internet: markus@gutschke.com
 *  My snail mail address is:
 *	Markus Gutschke
 *	3637 Fillmore Street #106
 *	San Francisco, CA 94123-1600
 *	USA
 *  If you like this software, I would appreciate if you sent me a postcard
 *  from your hometown. Under the terms of the GNU general public license
 *  you are free to include this program into (commercial) software
 *  distributions (e.g. putting it onto CD-ROM); nonetheless, I would really
 *  appreciate if you dropped me a short note (sending me a sample copy of
 *  your distribution would be even more appreciated!)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <signal.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <linux/vt.h>
#include <sys/wait.h>

#include "terminal.h"
#include "osd.h"
#include "rc.h"
#include "rcall.h"
#include "server.h"
#include "tv.h"
#include "tvcontrol.h"

int useVT = 0;
static int restoreVGA = -1;
static int myVT = -1;
static int myTTY = -1;
static int oldVT = -1;

/* VT-console switching code:
 *    shamelessly stolen from SVGALib :-)
 *    thanx to Bernd "Bernie" Meyer <root@wombat.hanse.de> for
 *    tracking down this information!
 */

#define SETSIG(sig,fun) \
        do {struct sigaction __siga; \
	memset(&__siga,0,sizeof(__siga)); \
	__siga.sa_handler = fun; \
	__siga.sa_flags = 0; \
	sigaction(sig,&__siga,NULL); \
	} while (0)

static void releaseVT(int sig)
{
  extern void termRefresh(void);

  UNUSED(sig);
  if (restoreVGA < 0)
    restoreVGA = !isVisible();
  setVGA(1);
  clearScreen(); termRefresh();
  if (RCSettings.muteOnSwitch)
    tmpMute();
  ioctl(myTTY,VT_RELDISP,1L);
  SETSIG(SIGUSR1,releaseVT);
  return;
} 

static void acquireVT(int sig)
{
  UNUSED(sig);
  clearok(curscr,1);
  initializeTerminal();
  /* this call hangs console switches... kernel bug? */
  /* ioctl(myTTY,VT_RELDISP,2L); */ /* ??? */
  setVGA(restoreVGA > 0);
  refreshMute();
  restoreVGA = -1;
  SETSIG(SIGUSR2,acquireVT);
  return;
}

static void takeVTControl(void)
{
  struct vt_mode VTmode;
  
  if (!ioctl(myTTY,VT_GETMODE,&VTmode)) {
    SETSIG(SIGUSR1,releaseVT);
    SETSIG(SIGUSR2,acquireVT);
    VTmode.mode = VT_PROCESS;
    VTmode.relsig = SIGUSR1;
    VTmode.acqsig = SIGUSR2;
    ioctl(myTTY,VT_SETMODE,&VTmode); }
  else /* This should not happen! */
    fprintf(stderr,"Warning! "
	    "Execute from VT in order to detect VT switches\n");
  return;
}

static void initSignals(int flags)
{
  static void contHandler(int signum);
  static void stopHandler(int signum);
  static void intHandler(int signum);
  static void winchHandler(int signum);
  static void chldHandler(int signum);

  if (flags & 1) signal(SIGTSTP,stopHandler);
  if (flags & 2) signal(SIGCONT,contHandler);
  if (flags & 4) signal(SIGINT,intHandler);
  if (flags & 8) signal(SIGTERM,intHandler);
  if (flags &16) signal(SIGWINCH,winchHandler);
  return;
}

static void changeScreenSize(void)
{
/*extern int del_curterm(TERMINAL *term);

  del_curterm(cur_term);
  setupterm(NULL,fileno(stdin),NULL);
  initializeTerminal(); */
  return;
}

int warp(int destVT)
{
  int    rc = -1;
  struct vt_stat vts;
  int stdTTY,warpTTY,warpVT;
  char warpDevice[20];

  if ((stdTTY = open("/dev/tty0",O_RDONLY|O_NOCTTY)) < 0 ||
      ioctl(stdTTY,VT_OPENQRY,&warpVT)) {
    return(rc); }
  sprintf(warpDevice,"/dev/tty%d",warpVT);
  if ((warpTTY = open(warpDevice,O_RDWR)) < 0)
    warpTTY = stdTTY;
  else
    close(stdTTY);
  if (destVT < 0)
    destVT = myVT;
  if (ioctl(myTTY,VT_GETSTATE,&vts) == 0) {
    rc = vts.v_active;
    if (destVT == rc) {
      close(warpTTY);
      return(rc); } }
  if (destVT > 0)
    ioctl(warpTTY,VT_ACTIVATE,destVT);
  close(warpTTY);
  return(rc);
}

void initializeTerminal(void)
{
  static void closeTerminal(void);
  static int initialized = 0;

  if (initialized)
    endwin();
  else {
    int  err;
    if (useVT >= 0) {
      if ((myTTY = 0),(myVT = useVT) <= 0 && ioctl(myTTY,VT_OPENQRY,&myVT) &&
	  ((myTTY = open("/dev/tty0",O_RDONLY|O_NOCTTY)) < 0 ||
	   ioctl(myTTY,VT_OPENQRY,&myVT))) {
	fprintf(stderr,"Warning! Could not allocate virtual console\n");
	myVT = useVT = -1; }
      else {
	struct vt_stat vts;
	char myDevice[20];
	useVT = myVT;
	sprintf(myDevice,"/dev/tty%d",myVT);
	if (myTTY >= 0) {
	  int tty = open(myDevice,O_RDWR);
	  if (tty >= 0) {
	    close(myTTY);
	    myTTY = tty; } }
	if (ioctl(myTTY,VT_GETSTATE,&vts) == 0)
	  oldVT = vts.v_active;
	if (freopen(myDevice,"r",stdin) == NULL) {
	  fprintf(stderr,"Could not redirect stdin to %s: %s\n",myDevice,strerror(errno));
	  exit(1); }
	if (freopen(myDevice,"a+",stdout) == NULL) {
	  fprintf(stderr,"Could not redirect stdout to %s: %s\n",myDevice,strerror(errno));
	  exit(1); } } }
    if (getenv("TERM") == NULL) {
      fprintf(stderr,"Environment variable TERM not set; assuming `vt100'\n");
      putenv("TERM=vt100"); }
    if (setupterm(NULL,fileno(stdin),&err) < 1) {
      fprintf(stderr,"Environment variable TERM is not valid; resetting to "
	             "`vt100'\n");
      putenv("TERM=vt100"); }
    if (initscr() == NULL) {
      fprintf(stderr,"Cannot open curses screen\n");
      exit(1); }
    if (myVT > 0)
      takeVTControl();
    atexit(closeTerminal);
    initialized = 1; }
  curs_set(0);
  leaveok(stdscr,1);
  cbreak();
  noecho();
  nonl();
  clearok(curscr,1);
  meta(stdscr,1);
  keypad(stdscr,1);
  scrollok(stdscr,0);
  refresh();
  initSignals(~0);
  return;
}

static void closeTerminal(void)
{
  if (!rcall(NULL,NULL))
    saveSettings();
  initSignals(~0);
  if (myVT > 0) {
    ioctl(myTTY,VT_ACTIVATE,oldVT > 0 ? oldVT : 1);
    ioctl(myTTY,VT_WAITACTIVE,oldVT > 0 ? oldVT : 1); }
  clearok(curscr,1);
  refresh();
  endwin();
  signal(SIGCHLD,SIG_IGN);
  return;
}

static void stopHandler(int signum)
{
  UNUSED(signum);
  endwin();
  if (restoreVGA < 0)
    restoreVGA = !isVisible();
  setVGA(1);
  initSignals(~1);
  raise(SIGTSTP);
  return;
}

static void contHandler(int signum)
{
  extern void termRefresh(void);

  UNUSED(signum);
  clearok(curscr,1);
  initializeTerminal();
  if (myVT > 0)
    ioctl(myTTY,VT_ACTIVATE,myVT);
  setVGA(restoreVGA > 0);
  restoreVGA = -1;
  clearScreen(); termRefresh();
  initSignals(~0);
  return;
}

static void intHandler(int signum)
{
  UNUSED(signum);
  exit(1);
}

static void winchHandler(int signum)
{
  UNUSED(signum);
  changeScreenSize();
  initSignals(~0);
  return;
}

int getKey(void)
{
  extern void checkTvTimeout(void);
  int ch;

  initSignals(~0);
  checkTvTimeout();
  if (waitReadable(fileno(stdin),5)) {
    ch = getch();
    return(ch >= 0 && ch < 256 ? toupper(ch) : ch); }
  else
    return(ERR);
}

